/*
 * Copyright (c) 2008-2009,2018 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/queue.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ftdi.h>
#include <usb.h>

#include "rmp/rmpLib.h"
#include "rmpreg.h"

static u_int8_t rmpChecksum(u_int8_t *);
static int rmpDecodeMessage(u_int8_t *, RMP_DATA_STR *);

/*----------------------------------------------------------------------*/

/**
 ** Open and initialize all RMP devices found on the USB busses
 **
 ** return: 0 or a negative error code
 **/
int
rmpOpenAll(struct RMP_DEV_STR ***rmps)
{
	struct ftdi_context *ctx;
	struct ftdi_device_list *list, *l;
	struct RMP_DEV_STR *dev;
	int i, n, err, retval;
	
	ctx = ftdi_new();
	if (ctx == NULL)
		return S_rmpLib_NO_MEMORY;

	n = ftdi_usb_find_all(ctx, &list, USB_VENDOR_SEGWAY, USB_PRODUCT_RMP);
	if (n <= 0) {
		fprintf(stderr, "ftdi_usb_find_all: %s\n",
			ftdi_get_error_string(ctx));
		ftdi_free(ctx);
		return S_rmpLib_FTDI_SYSTEM_ERROR;
	}
	*rmps = (struct RMP_DEV_STR **)calloc(n, sizeof(struct RMP_DEV_STR *));
	if (*rmps == NULL) {
		ftdi_free(ctx);
		return S_rmpLib_NO_MEMORY;
	}
	ftdi_free(ctx);
	retval = 0;
	for (l = list, i = 0; i < n; l = l->next, i++) {
		dev = (struct RMP_DEV_STR *)calloc(1, sizeof(struct RMP_DEV_STR));
		if (dev == NULL) {
			retval = S_rmpLib_NO_MEMORY;
			break;
		}
		pthread_mutex_init(&dev->mutex, NULL);
		pthread_mutex_lock(&dev->mutex);
		dev->state = (struct RMP_DATA_STR *)
		    calloc(1, sizeof(struct RMP_DATA_STR));
		if (dev->state == NULL) {
			retval = S_rmpLib_NO_MEMORY;
			free(dev);
			break;
		}
		pthread_mutex_unlock(&dev->mutex);
		(*rmps)[i] = dev;
		

		SIMPLEQ_INIT(&dev->paquetQueue);
		ctx = ftdi_new();
		if (ctx == NULL) {
			retval = S_rmpLib_NO_MEMORY;
			break;
		}
		dev->ftdi_context = ctx;
		err = ftdi_usb_open_dev(ctx, l->dev);
		if (err < 0) {
			fprintf(stderr, "ftdi_udb_open_dev: %s\n",
				ftdi_get_error_string(ctx));
			retval = S_rmpLib_FTDI_SYSTEM_ERROR;
			break;
		}
			
		err = ftdi_set_baudrate(ctx, 460800);
		if (err < 0) {
			fprintf(stderr, "ftdi_set_baud_rate: %s\n",
			    ftdi_get_error_string(ctx));
			retval = S_rmpLib_FTDI_SYSTEM_ERROR;
			break;
		}
#if 0
		err = ftdi_read_data_set_chunksize(ctx, 
		    RMP_USB_PACKET_SIZE*100);
		if (err < 0) {
			fprintf(stderr, "ftdi_read_data_set_chunksize: %s\n",
			    ftdi_get_error_string(ctx));
			retval = S_rmpLib_FTDI_SYSTEM_ERROR;
		}
#endif
		err = ftdi_write_data_set_chunksize(ctx, 
		    RMP_USB_PACKET_SIZE);
		if (err < 0) {
			fprintf(stderr, "ftdi_write_data_set_chunksize: %s\n",
			    ftdi_get_error_string(ctx));
			retval = S_rmpLib_FTDI_SYSTEM_ERROR;
		}
		err = ftdi_set_latency_timer(ctx, 1);
		if (err < 0) {
			fprintf(stderr, "ftdi_set_latency: %s\n",
			    ftdi_get_error_string(ctx));
			retval = S_rmpLib_FTDI_SYSTEM_ERROR;
		}
		ftdi_usb_purge_buffers(ctx);
		dev->magic = RMP_MAGIC;
	}
	ftdi_list_free(&list);
	if (retval < 0) {
		for (i = 0; i < n; i++) {
			dev = (*rmps)[i];
			if (dev == NULL)
				break;
			if (dev->state != NULL) 
				free(dev->state);
			if (dev->ftdi_context != NULL) 
				ftdi_free(dev->ftdi_context);
			free(dev);
		}
		return retval;
	}

	return n;
}

/*----------------------------------------------------------------------*/

/**
 ** Open and initialize one particular RMP device
 **
 ** Input: 
 **   serial : string representing the USB serial number of the device
 **
 ** Returns: NULL if an error occured 
 **          or a pointer to a new RMP_DEV structure
 **/
struct RMP_DEV_STR *
rmpOpen(char *serial)
{
	struct RMP_DEV_STR *dev;
	struct ftdi_context *ctx;
	int err;

	dev = (struct RMP_DEV_STR *)malloc(sizeof(struct RMP_DEV_STR));
	if (dev == NULL)
		return NULL;

	pthread_mutex_init(&dev->mutex, NULL);
	pthread_mutex_lock(&dev->mutex);
	dev->state = (struct RMP_DATA_STR *)
	    calloc(sizeof(struct RMP_DATA_STR), 1);
	pthread_mutex_unlock(&dev->mutex);
	if (dev->state == NULL) {
		free(dev);
		return NULL;
	}
	SIMPLEQ_INIT(&dev->paquetQueue);
	ctx = ftdi_new();
	if (ctx == NULL) {
		free(dev->state);
		free(dev);
		return NULL;
	}
	dev->ftdi_context = ctx;
	err = ftdi_usb_open_desc(ctx, USB_VENDOR_SEGWAY, USB_PRODUCT_RMP, 
	    NULL, serial);
	if (err < 0) {
		fprintf(stderr, "ftdi_udb_open_desc: %s\n",
		    ftdi_get_error_string(ctx));
		free(dev);
		return NULL;
	}
	err = ftdi_set_baudrate(ctx, 460800);
	if (err < 0) {
		fprintf(stderr, "ftdi_set_baud_rate: %s\n",
		    ftdi_get_error_string(ctx));
		ftdi_free(ctx);
		free(dev);
		return NULL;
	}
#if 0
	err = ftdi_read_data_set_chunksize(ctx, RMP_USB_PACKET_SIZE*100);
	if (err < 0) {
		fprintf(stderr, "ftdi_read_data_set_chunksize: %s\n",
		    ftdi_get_error_string(ctx));
		ftdi_free(ctx);
		free(dev);
		return NULL;
	}
#endif
	err = ftdi_set_latency_timer(ctx, 1);
	if (err < 0) {
		fprintf(stderr, "ftdi_set_latency: %s\n",
		    ftdi_get_error_string(ctx));
		ftdi_free(ctx);
		free(dev);
		return NULL;
	}
	ftdi_usb_purge_buffers(ctx);
	dev->magic = RMP_MAGIC;
	dev->serial = serial;
	dev->npackets = 0;
	return dev;
}

/*----------------------------------------------------------------------*/

/**
 ** Read data from a RMP device, storing it in the device's internal buffer
 **
 ** Input:
 **    dev: device to read from
 **    block: 0 -> non-blocking read  (return 0  if nothing to read)
 **           1 -> blocking read (waits until data available)
 **
 ** Return: the number of byte read, or a negative return code on error
 **/

static int
rmpReadData(struct RMP_DEV_STR *dev, int block)
{
	struct ftdi_context *ctx = dev->ftdi_context;
	int nread;

	if (dev->magic != RMP_MAGIC)
		return S_rmpLib_BAD_MAGIC;

	while (dev->readPtr < RMP_USB_PACKET_SIZE) {
		nread = ftdi_read_data(ctx, dev->readBuf + dev->readPtr,
		    RMP_USB_CHUNK_SIZE - dev->readPtr);
		/* printf("%s: read %d bytes %d %02x\n", dev->serial, 
		   nread, dev->readPtr, dev->readBuf[0]); */
		if (block == 0 && nread == 0)
			return 0;
		if (nread < 0)
			return S_rmpLib_USB_READ_ERROR;

		dev->readPtr += nread;
	}
	return nread;
}


/*----------------------------------------------------------------------*/
/**
 ** Read one packet of data from RMP device
 **
 ** Input:
 **   dev: device to read from
 **   block: 0 -> non blocking read (returns 0 if no data available)
 **          1 -> waits for a packet
 ** Output:
 **   buf contains one packet of data
 **
 ** Return: 1 if a packet was read
 **         0 if non-block and no data available
 **        <0 error - error code
 **/
int
rmpReadPacket(struct RMP_DEV_STR *dev, unsigned char *buf, int block)
{
	struct usb_message *n;
	int i, nread;

	if (dev->magic != RMP_MAGIC)
		return S_rmpLib_BAD_MAGIC;

	while (SIMPLEQ_EMPTY(&dev->paquetQueue)) {
		/* Get more data */
		nread = rmpReadData(dev, block);
		if (nread  < 0)
			return nread;
		if (nread == 0 && block == 0) 
			return 0;
		/* nread > 0 -> Find packets */
		i = 0;
		while (i <= dev->readPtr - RMP_USB_PACKET_SIZE) {
			if (dev->readBuf[i] == RMP_USB_MESSAGE_HEADER &&
			    rmpChecksum(dev->readBuf + i) 
			    == dev->readBuf[i+RMP_USB_PACKET_SIZE - 1]) {
				/* Good packet */
				n = malloc(sizeof(struct usb_message));
				memcpy(&n->data, dev->readBuf+i, RMP_USB_PACKET_SIZE);
				SIMPLEQ_INSERT_TAIL(&dev->paquetQueue,
				    n, next);
				dev->npackets++;
				i += RMP_USB_PACKET_SIZE;
			} else {
				i++;
			}
		}
		/* Copy the remaining bytes to the beginning of the buffer */
		memmove(dev->readBuf, dev->readBuf + i, 
		    dev->readPtr - i);
		dev->readPtr -= i;
		/* fprintf(stderr, "%s: read %d packets\n", dev->serial, 
		   dev->npackets); */
	}
	n = SIMPLEQ_FIRST(&dev->paquetQueue);
	memcpy(buf, n->data, RMP_USB_PACKET_SIZE);
	pthread_mutex_lock(&dev->mutex);
	rmpDecodeMessage(buf, dev->state);
	pthread_mutex_unlock(&dev->mutex);
	SIMPLEQ_REMOVE_HEAD(&dev->paquetQueue, next);
	free(n);	
	dev->npackets--;
	/* fprintf(stderr, "%s: remains %d packets\n", dev->serial, 
	   dev->npackets); */
	return 1;
}

/**
 ** Read as many packets as possible, and decode them
 **
 ** Input: dev - device to read from
 **
 ** Updates device's state
 **
 ** Return number of packets that were decoded
 **/       
int
rmpReadPackets(struct RMP_DEV_STR *dev)
{
	int i, nread;
	int npackets = 0;

	if (dev->magic != RMP_MAGIC)
		return S_rmpLib_BAD_MAGIC;

	while (npackets < 100) {
		/* Get more data */
		nread = rmpReadData(dev, 0);
		if (nread  < 0)
			return nread;
		if (nread == 0) {
			/* no data - mark it as invalid */
			pthread_mutex_lock(&dev->mutex);
			memset(dev->state, 0, sizeof(RMP_DATA_STR));
			pthread_mutex_unlock(&dev->mutex);
			break;
		}
		/* nread > 0 -> Find packets */
		i = 0;
		while (i <= dev->readPtr - RMP_USB_PACKET_SIZE) {
			if (dev->readBuf[i] == RMP_USB_MESSAGE_HEADER &&
			    rmpChecksum(dev->readBuf + i) 
			    == dev->readBuf[i+RMP_USB_PACKET_SIZE - 1]) {
				/* Good packet */
				pthread_mutex_lock(&dev->mutex);
				rmpDecodeMessage(dev->readBuf+i, dev->state);
				pthread_mutex_unlock(&dev->mutex);
				i += RMP_USB_PACKET_SIZE;
				npackets++;
			} else {
				i++;
			}
		}
		/* Copy the remaining bytes to the beginning of the buffer */
		memmove(dev->readBuf, dev->readBuf + i, 
		    dev->readPtr - i);
		dev->readPtr -= i;
	}
	return npackets;
}

/**
 ** Alternative version of rmpReadPacket() 
 **/
int
rmpReadPacket2(struct RMP_DEV_STR *dev, unsigned char *buf)
{
	struct ftdi_context *ctx = dev->ftdi_context;
	int i, nread, remaining;
	u_int16_t id;
	u_int8_t cksum;
	
	if (dev->magic != RMP_MAGIC)
		return S_rmpLib_BAD_MAGIC;

	for (;;) {
		nread = ftdi_read_data(ctx, buf, RMP_USB_PACKET_SIZE);
		if (nread < 0) {
			fprintf(stderr, "read error %d\n", nread);
			return S_rmpLib_USB_READ_ERROR;
		}
		remaining = RMP_USB_PACKET_SIZE - nread;
		/* look for start */
		for (i = 0; i < nread; i++) {
			if (buf[i] == RMP_USB_MESSAGE_HEADER) {
				remaining += RMP_USB_PACKET_SIZE - i;
				break;
			}
		}
		if (i != nread) {
			if (i != 0) {
				fprintf(stderr, "skipped %d bytes\n", i);
				memmove(buf, buf+i, nread);
			}
			break;
		}
	}
	do {
		nread = ftdi_read_data(ctx, 
		    buf+RMP_USB_PACKET_SIZE-remaining, remaining);
		if (nread < 0) {

			fprintf(stderr, "read error 2 %d\n", nread);
			return S_rmpLib_USB_READ_ERROR;
		}
		remaining -= nread;
	} while (remaining > 0);
	cksum = rmpChecksum(buf);
	id = ((buf[4] << 3) | ((buf[5] >> 5) & 7)) & 0xffff;
	printf("id %04x checksum %02x %02x\n", id, cksum, buf[17]);
	return 0;
}

/*----------------------------------------------------------------------*/

/**
 ** Send a command message to a RMP device
 **
 ** Input:
 **   dev - the device to send the command to
 **   buf - a message ready to be sent
 **
 ** Return: 0 or error code
 **/
int
rmpSendMessage(struct RMP_DEV_STR *dev, unsigned char *buf)
{
	struct ftdi_context *ctx = dev->ftdi_context;
	int n;
	
	if (dev->magic != RMP_MAGIC)
		return S_rmpLib_BAD_MAGIC;

	buf[RMP_USB_PACKET_SIZE - 1] = rmpChecksum(buf);
	n = ftdi_write_data(ctx, buf, RMP_USB_PACKET_SIZE);
	if (n < 0) {
	  fprintf(stderr, "error ftdi_write_data %d: %s\n", n, 
		  ftdi_get_error_string(ctx));
		return S_rmpLib_USB_WRITE_ERROR;
	}
	return 0;
}

/*----------------------------------------------------------------------*/

/**
 ** Send a speed command to a device
 **
 ** Input:
 **   dev - the device
 **   v - linear speed in RMP coordinates/s
 **   w - angular speed in RMP coordinates/s
 **
 ** Return: 0 or error code
 **/

int
rmpSendSpeed(struct RMP_DEV_STR *dev, short v, short w)
{
	short data[4];
	
	data[0] = v;
	data[1] = w;
	data[2] = data[3] = 0;
	return rmpSendCanMsg(dev, data);
}

/**
 ** Send a power Off command to a device 
 **
 ** Input:
 **   dev - the device
 **
 ** Return: 0 or error code
 **/
int
rmpPowerOff(struct RMP_DEV_STR *dev)
{
	short data[4];
	
	data[0] = data[1] = 0;
	data[2] = 16;
	data[3] = 3;
	return rmpSendCanMsg(dev, data);
}


/**
 ** Send a power On command to a device 
 **
 ** Input:
 **   dev - the device
 **
 ** Return: 0 or error code
 **/
int
rmpPowerOn(struct RMP_DEV_STR *dev)
{
	short data[4];
	
	data[0] = data[1] = 0;
	data[2] = 16;
	data[3] = 1;
	return rmpSendCanMsg(dev, data);
}

/*----------------------------------------------------------------------*/

/**
 ** Send an arbitrary CAN message to a device
 **
 ** Input:
 **   dev - the device
 **   data - 4 16bit values representing the message to send;
 **
 ** Return: 0 or error code
 **/
int
rmpSendCanMsg(struct RMP_DEV_STR *dev, short *data)
{
	unsigned char msg[RMP_USB_PACKET_SIZE];
	int i;

	msg[0] = RMP_USB_MESSAGE_HEADER;
	msg[1] = RMP_USB_COMMAND;
	msg[2] = 0;
	msg[3] = 0;
	msg[4] = 0;
	msg[5] = 0;
	msg[6] = 0x04;
	msg[7] = 0x13;
	msg[8] = 0;

	for (i = 0; i < 4; i++) {
		msg[9+2*i] = data[i] >> 8;
		msg[10+2*i] = data[i] & 0xff;
	}
	return rmpSendMessage(dev, msg);
}
/*----------------------------------------------------------------------*/

/**
 ** Close the connection to a device and free associated resources 
 **
 ** Input:
 **   dev - the device
 **/
void
rmpClose(struct RMP_DEV_STR *dev)
{
	if (dev->magic != RMP_MAGIC)
		return;

	free(dev->state);
	ftdi_usb_close(dev->ftdi_context);
	ftdi_free(dev->ftdi_context);
	free(dev);
}

/*----------------------------------------------------------------------*/

/**
 ** Decode a  message sent by the RMP device
 **
 ** Input:
 **    message - pointer to the message to decode
 **
 ** Input/Output
 **    rmp - pointer to the internal state struct of the RMP device structure
 **          updated
 **
 ** Return 0 or error code
 **/
 

#define CAN_WORD(i) (message[9+2*(i)] << 8 | message[10+2*(i)])
#define CAN_LONG(i) (message[11+4*(i)]<<24|message[12+4*(i)]<<16|\
	    message[9+4*(i)]<<8|message[10+4*(i)])

static int
rmpDecodeMessage(u_int8_t *message, RMP_DATA_STR *rmp)
{
	u_int16_t header;

	/* Verify that it is an RMP message */
	if (message[0] != RMP_USB_MESSAGE_HEADER)
		return S_rmpLib_BAD_MESSAGE;
	if (message[1] != RMP_USB_COMMAND)
		return S_rmpLib_BAD_MESSAGE;

	/* Extract header */
	header = ((message[4] << 3) | ((message[5] >> 5) & 7)) & 0xffff;
	/* Decode CAN message */
	switch (header) {
	case RMP_MONITOR_DATA_1:
		/* ?? */
		break;
	case RMP_MONITOR_DATA_2:
		rmp->pitch_angle = CAN_WORD(0);
		rmp->pitch_rate = CAN_WORD(1);
		rmp->roll_angle = CAN_WORD(2);
		rmp->roll_rate = CAN_WORD(3);
		break;
	case RMP_MONITOR_DATA_3:
		rmp->lw_velocity = CAN_WORD(0);
		rmp->rw_velocity = CAN_WORD(1);
		rmp->yaw_rate = CAN_WORD(2);
		rmp->servo_frames = CAN_WORD(3);
		break;
	case RMP_MONITOR_DATA_4:
		rmp->integrated_left_wheel = CAN_LONG(0);
		rmp->integrated_right_wheel = CAN_LONG(1);
		break;
	case RMP_MONITOR_DATA_5:
		rmp->integrated_fore_aft = CAN_LONG(0);
		rmp->integrated_yaw = CAN_LONG(1);
		break;
	case RMP_MONITOR_DATA_6:
		rmp->left_torque = CAN_WORD(0);
		rmp->right_torque = CAN_WORD(1);
		break;
	case RMP_MONITOR_DATA_7:
		rmp->operational_mode = CAN_WORD(0);
		rmp->controller_gain_schedule = CAN_WORD(1);
		rmp->ui_voltage = CAN_WORD(2);
		rmp->powerbase_voltage = CAN_WORD(3);
		break;
	case RMP_MONITOR_DATA_8:
		rmp->velocity_command = CAN_WORD(0);
		rmp->turn_command = CAN_WORD(1);
		break;
	default:
		return S_rmpLib_BAD_MESSAGE;
	}
	return 0;
}

/*----------------------------------------------------------------------*/

/**
 ** Return a pointer to the internal state data of device dev
 **/
struct RMP_DATA_STR *
rmpGetData(struct RMP_DEV_STR *dev)
{
	if (dev->magic != RMP_MAGIC)
		return NULL;
	return dev->state;
}

/*----------------------------------------------------------------------*/
/** 
 ** Return the address of the mutex
 **/
pthread_mutex_t *
rmpGetMutex(struct RMP_DEV_STR *dev)
{
	if (dev->magic != RMP_MAGIC)
		return NULL;
	return &dev->mutex;
}

/*----------------------------------------------------------------------*/

/**
 ** Display the internal state structure
 **/
void
rmpShowRawData(struct RMP_DEV_STR *dev)
{
	struct RMP_DATA_STR *rmp = dev->state;

	if (dev->magic != RMP_MAGIC) {
		printf("Invalid dev %p\n", dev);
		return;
	}

	printf("-------------------------------------\n");
	printf("pitch angle: %d\n", rmp->pitch_angle);
	printf("pitch rate:  %d\n", rmp->pitch_rate);
	printf("roll angle:  %d\n", rmp->roll_angle);
	printf("roll rate:   %d\n", rmp->roll_rate);
	printf("yaw rate:    %d\n", rmp->yaw_rate);
	printf("left wheel velocity:  %d\n", rmp->lw_velocity);
	printf("right wheel velocity: %d\n", rmp->rw_velocity);
	printf("servo frames: %d\n", rmp->servo_frames);
	printf("integrated left wheel:  %d\n", rmp->integrated_left_wheel);
	printf("integrated right wheel: %d\n", rmp->integrated_right_wheel);
	printf("integrated fore/aft:    %d\n", rmp->integrated_fore_aft);
	printf("integrated yaw:         %d\n", rmp->integrated_yaw);
	printf("left torque:  %d\n", rmp->left_torque);
	printf("right torque: %d\n", rmp->right_torque);
	printf("operational mode:  %d\n", rmp->operational_mode);
	printf("controller gain schedule: %d\n", 
	    rmp->controller_gain_schedule);
	printf("ui_voltage:        %d\n", rmp->ui_voltage);
	printf("powerbase voltage: %d\n", rmp->powerbase_voltage);
	printf("velocity command:  %d\n", rmp->velocity_command);
	printf("turn command:      %d\n", rmp->turn_command);
}

	
/*----------------------------------------------------------------------*/

/**
 ** Compute the checksum of a message 
 ** Message is always 18 bytes long 
 **/
static unsigned char
rmpChecksum(u_int8_t *message)
{
	unsigned short checksum = 0;
	unsigned short checksum_hi;
	int i;

	for (i = 0; i < 17; i++)
		checksum += (short)message[i];
	checksum_hi = (unsigned short)(checksum >> 8);
	checksum &= 0xff;
	checksum += checksum_hi;
	checksum_hi = (unsigned short)(checksum >> 8);
	checksum &= 0xff;
	checksum += checksum_hi;
	checksum = (~checksum + 1) & 0xff;
	return (unsigned char)checksum;
}

